% The main ISAT routine

% Supporting functions begin with "isat_"

function [y,A,scenario] = isat(y0,atol)

% ISAT Global Variables
global records n_records b_tree n_b_tree

% Global Variables for ISAT analysis
global errors

opt = 'pope';
%opt = 'hdgr';
%
% ================================================================================================
% Compute the absolute tolerance squared
atol_sq = atol' * atol;
% Compute the number of states
n = size(y0,1);

% Select two variables for the 2-D visualization
v1 = 2;
v2 = 3;
% Plot the query point
scatter(y0(v1,1),y0(v2,1),20,'r','c','filled');
title('The query point');

% Initialize ISAT variables the first time through
if isempty(n_records),
   isat_initialize(y0,atol);
   % Plot first ellipse
   eoa = records(1).eoa;
   a = eoa(v1,v1)/atol_sq;
   b = eoa(v2,v2)/atol_sq;
   c = 2*eoa(v1,v2)/atol_sq;
   [xp,yp] = plot_ellipse(a,b,c,y0(v1,1),y0(v2,1));
   plot(xp,yp,'k');
   scatter(y0(v1,1),y0(v2,1),30,'g','s','filled');
   title('Initialize ISAT');
end
% ================================================================================================
% Search the ISAT database for a near record (not necessarily the nearest)
[i_branch i_branch2] = isat_bt_search(y0);

% Assign the index to "records"
i_leaf = b_tree(i_branch).leaf;

% Plot the near record located with the binary tree search
start = records(i_leaf).start;
scatter(start(v1,1),start(v2,1),20,'r','c','filled');
title('Locate a near record');

% Compute dy
dy_initial = y0 - records(i_leaf).start;
% ================================================================================================
% Determine if the record is within the initial state region of accuracy
eps_sq = dy_initial' * records(i_leaf).eoa * dy_initial;
if eps_sq <= atol_sq,
   isat_retrieval = 'y';
else
   isat_retrieval = 'n';
end
% ================================================================================================
% Scenario #1: Retrieval
if isat_retrieval == 'y',
   % Estimate of the final states
	y = records(i_leaf).finish + records(i_leaf).sensitivity * dy_initial;
	% Estimate of the sensitivities
   A = records(i_leaf).sensitivity;
   % Estimates of the final states and sensitivities are returned
   scenario = 1;
   
   % Turn the query point and nearest record black
   scatter(y0(v1,1),y0(v2,1),30,'k','c','filled');
   title('ISAT retrieval');
   scatter(start(v1,1),start(v2,1),30,'g','s','filled');
   return;
end
% ================================================================================================
% Determine if the record is within the final state region of accuracy
y = isat_integrate(y0);
% "dy" is the difference between the final state that ISAT predicts and the integration result
dy = y - (records(i_leaf).finish + records(i_leaf).sensitivity * dy_initial);
%
%
% Can remove this later
i = size(errors,1) + 1;
errors(i,:) = dy';
%
%
if opt == 'pope',
% Pope's method is to sequentially test each final state to determine if the ISAT approximation is
%   within a "box" of accuracy
% Initialize "within_final_tol"
within_final_tol = 'y';
% Test each element individually
for i = 1:n,
   if abs(dy(i,1)) > atol(i,1),
      within_final_tol = 'n';
      break;
   end
end
end
%
%
if opt == 'hdgr',
% Hedengren's method is to simultaneous test the final state to determine if the ISAT approximation is
%   within a "spheroid" of accuracy.
within_final_tol = 'n';
dy_sq = dy'*dy;
if dy_sq <= (atol_sq / size(Ty0,1)),
   within_final_tol = 'y';
end
end

% ================================================================================================
% Scenario #2: Growth
if within_final_tol == 'y',
   dy0 = y0 - records(i_leaf).start;
   % Pope's method
   c = 1;
   %
   if opt == 'hdgr',
   % Hedengren's method to estimate an expanded ellipsoid of accuracy based on dy'*dy
   if dy_sq > 0
      c = min(sqrt(atol_sq / (size(Ty0,1) * dy_sq)), 1.1);
   else
      % In case dy_sq = 0, avoid divide by zero
      c = 1.1;
   end
   end
   
   isat_growth(c * dy0, i_leaf, atol);
   
   % Estimate of the sensitivities
   A = records(i_leaf).sensitivity;
   
   % Exact final states and estimated sensitivities are returned
   scenario = 2;
   
   % Turn the query point into a blue triangle
   scatter(y0(v1,1),y0(v2,1),30,'b','d','filled');
   scatter(start(v1,1),start(v2,1),30,'g','s','filled');
   
   eoa = records(i_leaf).eoa;
   center = records(i_leaf).start;
	a = eoa(v1,v1)/atol_sq;
	b = eoa(v2,v2)/atol_sq;
	c = 2*eoa(v1,v2)/atol_sq;
	[xp,yp] = plot_ellipse(a,b,c,center(v1,1),center(v2,1));
	plot(xp,yp,'b');
   title('ISAT growth');
   
   return;
end
% ================================================================================================
% Scenario #3: Addition
% Calculate Sensitivity Matrix
A = isat_sensitivities(y0);

% Add leaf to "b_tree" and a record to "records"
isat_add(y0,y,A,i_branch,atol);
% Exact final states and exact sensitivities are returned
scenario = 3;
% Turn the query point into a green square
scatter(y0(v1,1),y0(v2,1),30,'g','s','filled');

eoa = records(n_records).eoa;
a = eoa(v1,v1)/atol_sq;
b = eoa(v2,v2)/atol_sq;
c = 2*eoa(v1,v2)/atol_sq;
[xp,yp] = plot_ellipse(a,b,c,y0(v1,1),y0(v2,1));
plot(xp,yp,'k');

scatter(start(v1,1),start(v2,1),30,'g','s','filled');
title('ISAT addition');




